﻿using System;
using System.Collections.Generic;
using System.Text;

using Opc.Ua;
using Opc.Ua.Client;

using System.Threading;

namespace MDPoverUA
{
    class UaLogic
    {
        #region constructor
        // constructor
        public UaLogic(Delegate notificationDelegate, Delegate enableDelegate, string xmlPath, string applicationName)
        {
            mThrdMainForm = System.Threading.Thread.CurrentThread;
            this.notificationDelegate = notificationDelegate;
            this.enableDelegate = enableDelegate;
            this.myNodeIds = new List<MyNodeId>();
            this.nodeIds = new List<NodeId>();
            this.applicationName = applicationName;

            // load properties from given XML file
            this.loadXml(xmlPath);

            // try to connect
            Thread connectThread = new Thread(this.tryToConnect);
            connectThread.Start();
        }
        #endregion

        #region class variables
        // a small helper class
        class MyNodeId
        {
            public MyNodeId(string identifier, string namespaceString, string type)
            {
                this.identifier = identifier;
                this.namespaceString = namespaceString;
                this.type = type;
            }

            string identifier, namespaceString, type;

            public string Identifier
            {
                get { return identifier; }
            }

            public string NamespaceString
            {
                get { return namespaceString; }
            }

            public string Type
            {
                get { return type; }
            }
        }

        // delegates to interact with the main form
        Delegate notificationDelegate;
        Delegate enableDelegate;

        // obejcts for information of OPC UA connection
        string applicationName;
        string url;
        string policy;
        int interval;
        Browser browser;
        List<MyNodeId> myNodeIds;
        List<NodeId> nodeIds;
        Subscription subscription;
        bool isTestApplication;
        int testInterval;
        int keepAlive;
        bool isConnected = false;

        System.Threading.Thread mThrdMainForm = null;        

        #region public objects

        public bool IsTestApplication
        { get { return this.isTestApplication; } }

        public int Interval { get { return this.testInterval; } }

        public bool IsConnected { get { return this.isConnected; } }

        #endregion
        #endregion

        #region public methods
        // forwards disconnect command
        public bool Disconnect()
        {
            return disconnect(true);
        }

        // looks up and returns the NodeId of the given identifier
        public NodeId GetNodeId(string identifier)
        {
            foreach (NodeId nodeId in this.nodeIds)
                if (nodeId.Identifier.Equals(identifier))
                    return nodeId;

            return null;
        }

        // reads and returns the current value of the given Node
        public DataValue ReadValue(NodeId nodeId)
        {
             return this.browser.Session.ReadValue(nodeId);
        }

        // writes the given value to the given Node
        public StatusCode WriteValue(NodeId nodeId, object value)
        {
            // get Node for values Type information
            Node node = this.browser.Session.ReadNode(nodeId);

            // if it is not a variable return error code
            if (!node.NodeClass.Equals(NodeClass.Variable))
                return StatusCodes.BadNodeClassInvalid;

            // preparement for writing process
            WriteValueCollection nodesToWrite = new WriteValueCollection();
            WriteValue nodeToWrite = new WriteValue();
            nodeToWrite.AttributeId = Attributes.Value;
            nodeToWrite.NodeId = nodeId;

            // check if value is type of array and set the value
            if (((VariableNode)node).ValueRank.Equals(1))
            {
                string[] tempElements = value.ToString().Split(new char[] { ';' }, StringSplitOptions.RemoveEmptyEntries);
                System.Collections.ArrayList tempArray = new System.Collections.ArrayList();
                Type type = DataTypes.GetSystemType(
                    (NodeId)((VariableNode)node).DataType,
                    EncodeableFactory.GlobalFactory
                    );
                foreach (string tempElement in tempElements)
                    tempArray.Add(Convert.ChangeType(tempElement, type));
                nodeToWrite.Value = new DataValue(new Variant(tempArray.ToArray(type)));
            }
            else
            {
                nodeToWrite.Value = new DataValue(new Variant(Convert.ChangeType(
                    value,
                    DataTypes.GetSystemType(
                        (NodeId)((VariableNode)node).DataType,
                        EncodeableFactory.GlobalFactory
                        )
                    )));
            }
            nodesToWrite.Add(nodeToWrite);

            StatusCodeCollection results = new StatusCodeCollection();
            DiagnosticInfoCollection diagnosticInfos = new DiagnosticInfoCollection();

            // write process
            this.browser.Session.Write(
                null,
                nodesToWrite,
                out results,
                out diagnosticInfos
                );

            // return StatusCode
            return results[0];
        }

        #endregion

        #region private methods
        // loads properties from given XML file
        void loadXml(string xmlPath)
        {
            System.Xml.XmlDocument xmlDocument = new System.Xml.XmlDocument();
            xmlDocument.Load(xmlPath);

            foreach (System.Xml.XmlNode xmlNode in xmlDocument.ChildNodes[0])
            {
                switch (xmlNode.Name)
                {
                    case "Session":
                        // set Session properties
                        this.url = xmlNode.Attributes["Url"].Value;

                        this.policy = xmlNode.Attributes["Security"].Value;
                        this.interval = int.Parse(xmlNode.Attributes["RefreshInterval"].Value);
                        this.keepAlive = int.Parse(xmlNode.Attributes["KeepAlive"].Value);                        
                        break;

                    case "NodeId":
                        // create NodeIds
                        this.myNodeIds.Add(new MyNodeId(xmlNode.Attributes["Identifier"].Value, xmlNode.Attributes["Namespace"].Value, xmlNode.Attributes["Type"].Value));
                        break;

                    case "LongTermTestApp":
                        // set the switches to execute a long term test
                        isTestApplication = bool.Parse(xmlNode.Attributes["Active"].Value);
                        testInterval = int.Parse(xmlNode.Attributes["Interval"].Value);
                        break;
                }
            }
        }

        // looks up the given namespace string and returns the correct namespace index of the UA server
        ushort NamespaceStringToIndex(string namespaceString)
        {
            string[] namespaces = browser.Session.NamespaceUris.ToArray();
            for (int i = 0; i < namespaces.Length; i++)
                if (namespaceString.ToLower().Equals(namespaces[i].ToLower()))
                    return (ushort)i;
            return ushort.MaxValue;
        }

        // create a Subscription for each variable
        void createSubscriptions()
        {
            ushort namespaceIndex = ushort.MaxValue;
            string lastNamespace = string.Empty;

            foreach (MyNodeId myNode in myNodeIds)
            {
                if (namespaceIndex.Equals(ushort.MaxValue) || !lastNamespace.Equals(myNode.NamespaceString))
                {
                    lastNamespace = myNode.NamespaceString;
                    namespaceIndex = NamespaceStringToIndex(lastNamespace);
                }

                // check the type of NodeId and create and add the correct one
                switch (myNode.Type.ToLower())
                {
                    case "numeric":
                        nodeIds.Add(new NodeId(uint.Parse(myNode.Identifier), namespaceIndex));
                        break;

                    case "guid":
                        nodeIds.Add(new NodeId(new Guid(myNode.Identifier), namespaceIndex));
                        break;

                    default:
                        nodeIds.Add(new NodeId(myNode.Identifier, namespaceIndex));
                        break;
                }
            }

            // create a Subscription first
            this.subscription = new Subscription(this.browser.Session.DefaultSubscription);
            this.subscription.PublishingEnabled = true;
            this.subscription.PublishingInterval = this.interval;
            this.browser.Session.AddSubscription(this.subscription);
            this.subscription.Create();

            // finally append MonitoredItems
            foreach (NodeId nodeId in this.nodeIds)
            {
                MonitoredItem item = new MonitoredItem(subscription.DefaultItem);
                item.DisplayName = nodeId.Identifier.ToString();
                item.StartNodeId = nodeId;
                item.NodeClass = NodeClass.Variable;
                item.AttributeId = Attributes.Value;
                item.DiscardOldest = true;
                item.SamplingInterval = interval;
                // link with parent event later!!!
                item.Notification += new MonitoredItemNotificationEventHandler(item_Notification);

                this.subscription.AddItem(item);
            }

            this.subscription.ApplyChanges();
        }

        // delegates the main form to analyse the event
        void item_Notification(MonitoredItem monitoredItem, MonitoredItemNotificationEventArgs e)
        {
            lock (this)
            {
                if (browser != null
                    && browser.Session != null
                    && browser.Session.Connected)
                {
                    this.notificationDelegate.DynamicInvoke(monitoredItem, e);
                }
            }
        }

        #region (dis-)connect methods
        // connects to the given OPC UA server
        bool connect(string serverUrl, string policy)
        {
            // create the configuration.     
            ApplicationConfiguration configuration = this.createClientConfiguration();

            // create the endpoint description.
            EndpointDescription endpointDescription = this.createEndpointDescription(serverUrl, policy);

            // create the endpoint configuration (use the application configuration to provide default values).
            EndpointConfiguration endpointConfiguration = EndpointConfiguration.Create(configuration);

            // the default timeout for a requests sent using the channel.
            endpointConfiguration.OperationTimeout = 60000;

            // use the pure XML encoding on the wire.
            endpointConfiguration.UseBinaryEncoding = true;

            // create the endpoint.
            ConfiguredEndpoint endpoint = new ConfiguredEndpoint(null, endpointDescription, endpointConfiguration);

            // create the binding factory.
            BindingFactory bindingFactory = BindingFactory.Create(configuration);

            // update endpoint description using the discovery endpoint.
            if (endpoint.UpdateBeforeConnect)
            {
                endpoint.UpdateFromServer(bindingFactory);

                endpointDescription = endpoint.Description;
                endpointConfiguration = endpoint.Configuration;
            }

            System.Security.Cryptography.X509Certificates.X509Certificate2 clientCertificate = configuration.SecurityConfiguration.ApplicationCertificate.Find();

            // set up a callback to handle certificate validation errors.
            configuration.CertificateValidator.CertificateValidation += new CertificateValidationEventHandler(CertificateValidator_CertificateValidation);

            // Initialize the channel which will be created with the server.
            SessionChannel channel = SessionChannel.Create(
                configuration,
                endpointDescription,
                endpointConfiguration,
                bindingFactory,
                clientCertificate,
                null);

            // Wrap the channel with the session object.
            // This call will fail if the server does not trust the client certificate.
            Session session = new Session(channel, configuration, endpoint);
            session.ReturnDiagnostics = DiagnosticsMasks.All;

            // create the session. This actually connects to the server.            
            session.Open(this.applicationName + " Session", null);

            session.KeepAliveInterval = keepAlive;
            session.KeepAlive += new KeepAliveEventHandler(session_KeepAlive);

            // create the Browser object
            this.browser = new Browser(session);
            this.browser.ReferenceTypeId = ReferenceTypes.HierarchicalReferences;

            isConnected = true;
            return true;
        }

        // check if there is still a connection to the given UA server
        void session_KeepAlive(Session session, KeepAliveEventArgs e)
        {
            // if there is a problem with the connection to the UA server
            if (session.KeepAliveStopped)
            {
                // disable controls in the form
                try
                {
                    this.enableDelegate.DynamicInvoke(false);
                }
                catch { }

                // do a clean disconnect
                try { disconnect(false); }
                catch { }

                // try a reconnect until the process is successfull (in case of a network adapter release)
                while (true)
                {
                    try
                    {
                        // try to connect and restore the subscriptions
                        if (connect(url, policy))
                        {
                            System.Threading.Thread.Sleep(interval);
                            // subscribe all variables if connected
                            this.createSubscriptions();

                            // enable the controls of the form
                            try { this.enableDelegate.DynamicInvoke(true); }
                            catch { }
                            return;
                        }
                        // if there is no connect wait half a second
                        else
                        { System.Threading.Thread.Sleep(500); }
                    }
                    // in case of an error wait half a second
                    catch { System.Threading.Thread.Sleep(500); }
                }
            }
        }

        // called in case of a timeout and tries to reconnect
        private void tryToConnect()
        {
            while (true)
            {
                try
                {
                    if (connect(url, policy))
                    {
                        System.Threading.Thread.Sleep(interval);
                        // subscribe all variables if connected
                        this.createSubscriptions();

                        try { this.enableDelegate.DynamicInvoke(true); }
                        catch { }
                        return;
                    }
                    else
                    { System.Threading.Thread.Sleep(500); }
                }
                catch { System.Threading.Thread.Sleep(500); }
            }
        }

        // disconnects from the OPC UA server currently connected to
        bool disconnect(bool closing)
        {
            isConnected = false;

            bool bDoDisconnect = false;
            Subscription subscriptionTmp = null;
            Browser browserTmp = null;

            // check if the program is not terminated
            if (!closing)
                lock (this)
                {
                    if (this.browser.Session != null)
                    {
                        bDoDisconnect = true;
                        subscriptionTmp = this.subscription;
                        browserTmp = this.browser;

                        this.subscription = null;
                        this.browser = null;
                    }
                }
            else
            {
                object obj = new object();
                lock (obj)
                {
                    if (this.browser.Session != null)
                    {
                        bDoDisconnect = true;
                        subscriptionTmp = this.subscription;
                        browserTmp = this.browser;

                        this.subscription = null;
                        this.browser = null;
                    }
                }
            }

            if (bDoDisconnect)
            {
                if (browserTmp.Session != null
                    && browserTmp.Session.Connected)
                {
                    subscriptionTmp.DeleteItems();
                    subscriptionTmp.Dispose();
                    subscriptionTmp = null;

                    // disconnect from the OPC UA server
                    browserTmp.Session.Close();

                    // release ressources
                    browserTmp.Session.Dispose();
                    browserTmp = null;

                    return true;
                }
                else
                { return false; }
            }
            else
            { return false; }
        }        

        #region connect help methods

        // handles untrusted certificates
        void CertificateValidator_CertificateValidation(CertificateValidator validator, CertificateValidationEventArgs e)
        {
            e.Accept = true;
        }

        // creates a configuration of this application
        ApplicationConfiguration createClientConfiguration()
        {
            // The application configuration can be loaded from any file.
            // ApplicationConfiguration.Load() method loads configuration by looking up a file path in the App.config.
            // This approach allows applications to share configuration files and to update them.
            ApplicationConfiguration configuration = new ApplicationConfiguration();

            // Step 1 - Specify the server identity.
            configuration.ApplicationName = this.applicationName;
            configuration.ApplicationType = ApplicationType.Client;
            configuration.ApplicationUri = "http://localhost/VendorId/ApplicationId/InstanceId";
            configuration.ProductUri = "http://VendorId/ProductId/VersionId";

            configuration.SecurityConfiguration = new SecurityConfiguration();

            // Step 2 - Specify the server's application instance certificate.

            // Application instance certificates must be placed in a windows certficate store because that is 
            // the best way to protect the private key. Certificates in a store are identified with 4 parameters:
            // StoreLocation, StoreName, SubjectName and Thumbprint.
            //
            // In this example the following values are used:
            // 
            //   LocalMachine    - use the machine wide certificate store.
            //   Personal        - use the store for individual certificates.
            //   ApplicationName - use the application name as a search key.   

            configuration.SecurityConfiguration.ApplicationCertificate = new CertificateIdentifier();
            configuration.SecurityConfiguration.ApplicationCertificate.StoreType = CertificateStoreType.Windows;
            configuration.SecurityConfiguration.ApplicationCertificate.StorePath = "LocalMachine\\My";
            configuration.SecurityConfiguration.ApplicationCertificate.SubjectName = configuration.ApplicationName;

            // trust all applications installed on the same machine.
            configuration.SecurityConfiguration.TrustedPeerCertificates.StoreType = CertificateStoreType.Windows;
            configuration.SecurityConfiguration.TrustedPeerCertificates.StorePath = "LocalMachine\\My";

            // find the certificate in the store.
            System.Security.Cryptography.X509Certificates.X509Certificate2 clientCertificate = configuration.SecurityConfiguration.ApplicationCertificate.Find(true);

            // create a new certificate if one not found.
            if (clientCertificate == null)
            {
                // this code would normally be called as part of the installer - called here to illustrate.
                // create a new certificate an place it in the LocalMachine/Personal store.
                clientCertificate = CertificateFactory.CreateCertificate(
                    configuration.SecurityConfiguration.ApplicationCertificate.StoreType,
                    configuration.SecurityConfiguration.ApplicationCertificate.StorePath,
                    configuration.ApplicationUri,
                    configuration.ApplicationName,
                    null,
                    null,
                    1024,
                    120);
            }

            // Step 3 - Specify the supported transport configurations.

            // The SDK requires Binding objects which are sub-types of Opc.Ua.Bindings.BaseBinding
            // These two lines add support for SOAP/HTTP w/ WS-* and UA-TCP. Support for other protocols 
            // such as .NET TCP can be added but they would not be considered interoperable across different vendors. 
            // Only one binding per URL scheme is allowed.
            configuration.TransportConfigurations.Add(new TransportConfiguration(Utils.UriSchemeOpcTcp, typeof(Opc.Ua.Bindings.UaTcpBinding)));
            configuration.TransportConfigurations.Add(new TransportConfiguration(Utils.UriSchemeHttp, typeof(Opc.Ua.Bindings.UaSoapXmlBinding)));

            // Step 4 - Specify the supported transport quotas.

            // The transport quotas are used to set limits on the contents of messages and are
            // used to protect against DOS attacks and rogue clients. They should be set to
            // reasonable values.
            configuration.TransportQuotas = new TransportQuotas();
            configuration.TransportQuotas.OperationTimeout = 60000;

            configuration.ServerConfiguration = new ServerConfiguration();

            // Step 5 - Specify the client specific configuration.
            configuration.ClientConfiguration = new ClientConfiguration();
            configuration.ClientConfiguration.DefaultSessionTimeout = 30000;

            // Step 6 - Validate the configuration.

            // This step checks if the configuration is consistent and assigns a few internal variables
            // that are used by the SDK. This is called automatically if the configuration is loaded from
            // a file using the ApplicationConfiguration.Load() method.          
            configuration.Validate(ApplicationType.Client);

            return configuration;
        }

        // specifies information for the connection to the OPC UA server
        EndpointDescription createEndpointDescription(string serverUrl, string policy)
        {
            // create the endpoint description.
            EndpointDescription endpointDescription = new EndpointDescription();

            endpointDescription.EndpointUrl = serverUrl;

            // specify the security policy to use.
            if (policy.Equals("Basic128Rsa15"))
            {
                endpointDescription.SecurityPolicyUri = SecurityPolicies.Basic128Rsa15;
                endpointDescription.SecurityMode = MessageSecurityMode.SignAndEncrypt;
            }
            else if (policy.Equals("None"))
            {
                endpointDescription.SecurityPolicyUri = SecurityPolicies.None;
                endpointDescription.SecurityMode = MessageSecurityMode.None;
            }

            // specify the transport profile.
            endpointDescription.TransportProfileUri = Profiles.WsHttpXmlOrBinaryTransport;

            return endpointDescription;
        }

        #endregion
        #endregion

        #endregion

    }
}
